@@ -101,6 +101,12 @@ create_gui (XfcePanelPlugin *plugin)
     if ((base->nr_cores = init_cpu_data (&base->cpu_data)) == 0)
         fprintf (stderr,"Cannot init cpu data !\n");
 
+    /* Read CPU data twice in order to initialize
+     * cpu_data[].previous_used and cpu_data[].previous_total
+     * with the current HWMs. HWM = High Water Mark. */
+    read_cpu_data (base->cpu_data, base->nr_cores);
+    read_cpu_data (base->cpu_data, base->nr_cores);
+
     base->topology = read_topology ();
 
     base->plugin = plugin;
@@ -154,7 +160,7 @@ about_cb (XfcePanelPlugin *plugin, CPUGraph *base)
         "program-name", PACKAGE_NAME,
         "comments", _("Graphical representation of the CPU load"),
         "website", "https://docs.xfce.org/panel-plugins/xfce4-cpugraph-plugin",
-        "copyright", _("Copyright (c) 2003-2020\n"),
+        "copyright", _("Copyright (c) 2003-2021\n"),
         "authors", auth, NULL);
 
     if (icon)
@@ -212,7 +218,7 @@ shutdown (XfcePanelPlugin *plugin, CPUGraph *base)
     gtk_widget_destroy (base->tooltip_text);
     if (base->timeout_id)
         g_source_remove (base->timeout_id);
-    g_free (base->history);
+    g_free (base->history.data);
     g_free (base->command);
     g_free (base);
 }
@@ -228,13 +234,55 @@ delete_bars (CPUGraph *base)
     }
 }
 
+static void
+clear_history (CPUGraph *base)
+{
+    gssize i;
+    for (i = 0; i < base->history.cap_pow2; i++)
+        base->history.data[i] = (CpuLoad) {};
+}
+
+static void
+resize_history (CPUGraph *base, gssize history_size)
+{
+    const guint fastest = get_update_interval_ms (RATE_FASTEST);
+    const guint slowest = get_update_interval_ms (RATE_SLOWEST);
+    gssize cap_pow2, old_cap_pow2, old_mask, old_offset;
+    CpuLoad *old_data;
+
+    old_cap_pow2 = base->history.cap_pow2;
+    old_data = base->history.data;
+    old_mask = base->history.mask;
+    old_offset = base->history.offset;
+
+    cap_pow2 = 1;
+    while (cap_pow2 < 128 * slowest / fastest)
+        cap_pow2 <<= 1;
+    while (cap_pow2 < history_size * slowest / fastest)
+        cap_pow2 <<= 1;
+
+    if (cap_pow2 != old_cap_pow2)
+    {
+        gssize i;
+        base->history.cap_pow2 = cap_pow2;
+        base->history.data = (CpuLoad*) g_malloc0 (cap_pow2 * sizeof (CpuLoad));
+        base->history.mask = cap_pow2 - 1;
+        base->history.offset = 0;
+        for (i = 0; i < old_cap_pow2 && i < cap_pow2; i++)
+            base->history.data[i] = old_data[(old_offset + i) & old_mask];
+        g_free (old_data);
+    }
+
+    base->history.size = history_size;
+}
+
 static gboolean
 size_cb (XfcePanelPlugin *plugin, guint size, CPUGraph *base)
 {
-    gint frame_h, frame_v, history;
+    gint frame_h, frame_v;
+    gssize history;
     GtkOrientation orientation;
     gint shadow_width = base->has_frame ? 2*1 : 0;
-    gint i;
 
     orientation = xfce_panel_plugin_get_orientation (plugin);
 
@@ -251,18 +299,16 @@ size_cb (XfcePanelPlugin *plugin, guint size, CPUGraph *base)
         history = size;
     }
 
+    if (history > base->history.cap_pow2 || base->non_linear)
+        resize_history (base, history);
+    else
+        base->history.size = history;
+
     gtk_widget_set_size_request (GTK_WIDGET (base->frame_widget), frame_h, frame_v);
-
-    base->history = (gfloat *) g_realloc (base->history, history * sizeof (gfloat));
-    for (i = base->history_size; i < history; i++)
-        base->history[i] = 0;
-    base->history_size = history;
-
     if (base->has_bars) {
         base->bars.orientation = orientation;
         set_bars_size (base);
     }
-
     set_border (base, base->has_border);
 
     return TRUE;
@@ -436,8 +482,7 @@ detect_smt_issues (CPUGraph *base)
                                             /* Decrease combined_usage by load_to_move */
                                             for (j = topo->cores[core].num_logical_cpus; load_to_move > 0 && j != 0;)
                                             {
-                                                j--;
-                                                guint cpu = topo->cores[core].logical_cpus[j];
+                                                guint cpu = topo->cores[core].logical_cpus[--j];
                                                 if (G_LIKELY (cpu < base->nr_cores))
                                                 {
                                                     if (optimal_load[cpu] >= load_to_move)
@@ -525,30 +570,40 @@ update_cb (gpointer user_data)
 
     if (base->tracked_core > base->nr_cores)
         base->cpu_data[0].load = 0;
-    else if (base->tracked_core != 0)
+    else if (base->tracked_core != 0 && G_LIKELY (base->tracked_core < base->nr_cores + 1))
         base->cpu_data[0].load = base->cpu_data[base->tracked_core].load;
 
     if (base->mode != MODE_DISABLED)
     {
-        /* Update the history and draw the graph */
+        CpuLoad load;
+
+        /* Update the history */
         if (base->non_linear)
         {
-            gssize i = base->history_size - 1;
-            while (i > 0)
+            const gssize mask = base->history.mask;
+            const gssize offset = base->history.offset;
+            gssize i;
+            for (i = base->history.size - 1; i > 0; i--)
             {
                 const gfloat scale = 256.0f;
                 gfloat a, b, factor;
-                a = base->history[i] * scale, b = base->history[i-1] * scale;
+                a = base->history.data[(offset+i) & mask].value * scale;
+                b = base->history.data[(offset+i-1) & mask].value * scale;
                 if (a < b) a++;
                 factor = (i * 2);
-                base->history[i--] = (a * (factor-1) + b) / factor / scale;
+                base->history.data[(offset+i) & mask].timestamp = 0;
+                base->history.data[(offset+i) & mask].value = (a * (factor-1) + b) / factor / scale;
             }
         }
-        else {
-            memmove (base->history + 1 , base->history, (base->history_size - 1) * sizeof (*base->history));
+        else
+        {
+            base->history.offset = (base->history.offset - 1) & base->history.mask;
         }
-        base->history[0] = base->cpu_data[0].load;
+        load.timestamp = g_get_real_time ();
+        load.value = base->cpu_data[0].load;
+        base->history.data[base->history.offset] = load;
 
+        /* Draw the history graph */
         gtk_widget_queue_draw (base->draw_area);
     }
 
@@ -719,6 +774,31 @@ command_cb (GtkWidget *w, GdkEventButton *event, CPUGraph *base)
     return FALSE;
 }
 
+/**
+ * get_update_interval_ms:
+ *
+ * Returns: update interval in milliseconds.
+ */
+guint
+get_update_interval_ms (CPUGraphUpdateRate rate)
+{
+    switch (rate)
+    {
+        case RATE_FASTEST:
+            return 250;
+        case RATE_FAST:
+            return 500;
+        case RATE_NORMAL:
+            return 750;
+        case RATE_SLOW:
+            return 1000;
+        case RATE_SLOWEST:
+            return 3000;
+        default:
+            return 750;
+    }
+}
+
 void
 set_startup_notification (CPUGraph *base, gboolean startup_notification)
 {
@@ -783,7 +863,11 @@ set_frame (CPUGraph *base, gboolean frame)
 void
 set_nonlinear_time (CPUGraph *base, gboolean nonlinear)
 {
-    base->non_linear = nonlinear;
+    if (base->non_linear != nonlinear)
+    {
+        base->non_linear = nonlinear;
+        clear_history (base);
+    }
 }
 
 void
@@ -793,33 +877,26 @@ set_smt (CPUGraph *base, gboolean highlight_smt)
 }
 
 void
-set_update_rate (CPUGraph *base, guint rate)
+set_update_rate (CPUGraph *base, CPUGraphUpdateRate rate)
 {
-    guint update;
+    gboolean change = (base->update_interval != rate);
+    gboolean init = (base->timeout_id == 0);
 
-    base->update_interval = rate;
-
-    if (base->timeout_id)
-        g_source_remove (base->timeout_id);
-
-    switch (base->update_interval)
+    if (change || init)
     {
-        case 0:
-            update = 250;
-            break;
-        case 1:
-            update = 500;
-            break;
-        case 2:
-            update = 750;
-            break;
-        case 3:
-            update = 1000;
-            break;
-        default:
-            update = 3000;
+        guint interval = get_update_interval_ms (rate);
+
+        base->update_interval = rate;
+        if (base->timeout_id)
+            g_source_remove (base->timeout_id);
+        base->timeout_id = g_timeout_add (interval, update_cb, base);
+
+        if (change && !init)
+        {
+            if (base->mode != MODE_DISABLED)
+                gtk_widget_queue_draw (base->draw_area);
+        }
     }
-    base->timeout_id = g_timeout_add (update, update_cb, base);
 }
 
 void
@@ -844,8 +921,7 @@ set_mode (CPUGraph *base, CPUGraphMode mode)
     {
         /* Hide graph and clear history */
         gtk_widget_hide (base->frame_widget);
-        for (gint i = 0; i < base->history_size; i++)
-            base->history[i] = 0;
+        clear_history (base);
     }
     else
     {
