@@ -36,7 +36,8 @@
 #define HIGHLIGHT_SMT_BY_DEFAULT FALSE
 #define MAX_LOAD_THRESHOLD 0.2
 
-typedef enum {
+typedef enum
+{
     MODE_DISABLED = -1,
     MODE_NORMAL = 0,
     MODE_LED = 1,
@@ -44,9 +45,20 @@ typedef enum {
     MODE_GRID = 3,
 } CPUGraphMode;
 
+/* Number of milliseconds between updates */
+typedef enum
+{
+    RATE_FASTEST = 0,
+    RATE_FAST = 1,
+    RATE_NORMAL = 2,
+    RATE_SLOW = 3,
+    RATE_SLOWEST = 4,
+} CPUGraphUpdateRate;
+
 enum { NUM_COLORS = 6 };
 
-typedef enum {
+typedef enum
+{
     BG_COLOR = 0,
     FG_COLOR1 = 1,
     FG_COLOR2 = 2,
@@ -55,6 +67,12 @@ typedef enum {
     SMT_ISSUES_COLOR = 5, /* NUM_COLORS-1 */
 } CPUGraphColorNumber;
 
+typedef struct
+{
+    gint64 timestamp; /* Microseconds since 1970-01-01 UTC, or zero */
+    gfloat value;     /* Range: from 0.0 to 1.0 */
+} CpuLoad;
+
 typedef struct
 {
     /* GUI components */
@@ -72,7 +90,7 @@ typedef struct
     GtkWidget *tooltip_text;
 
     /* Settings */
-    guint update_interval; /* Number of ms between updates. */
+    CPUGraphUpdateRate update_interval;
     gboolean non_linear;
     guint size;
     CPUGraphMode mode;
@@ -92,13 +110,19 @@ typedef struct
     /* Runtime data */
     guint nr_cores;
     guint timeout_id;
-    gfloat *history;
-    gssize history_size;
+    struct {
+        gssize cap_pow2;  /* Capacity. A power of 2. */
+        gssize size;      /* size <= cap_pow2 */
+        gssize mask;      /* Equals to (cap_pow2 - 1) */
+        gssize offset;    /* Circular buffer position. Range: from 0 to (cap_pow2 - 1) */
+        CpuLoad *data;    /* Circular buffer */
+    } history;
     CpuData *cpu_data;
     Topology *topology;
     CpuStats stats;
 } CPUGraph;
 
+guint get_update_interval_ms (CPUGraphUpdateRate rate);
 void set_bars (CPUGraph * base, gboolean bars);
 void set_border (CPUGraph *base, gboolean border);
 void set_color (CPUGraph *base, guint number, GdkRGBA color);
@@ -113,6 +137,6 @@ void set_size (CPUGraph *base, guint width);
 void set_smt (CPUGraph * base, gboolean highlight_smt);
 void set_startup_notification (CPUGraph *base, gboolean startup_notification);
 void set_tracked_core (CPUGraph *base, guint core);
-void set_update_rate (CPUGraph *base, guint rate);
+void set_update_rate (CPUGraph *base, CPUGraphUpdateRate rate);
 
 #endif /* !_XFCE_CPU_H_ */
